# -*- coding: utf-8 -*-
# Author: OmenApps. https://omenapps.com
import inspect

from django.apps import apps as django_apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connection
from django_extensions.management.color import color_style
from django_extensions.management.utils import signalcommand

TAB = "        "
HALFTAB = "    "


class Command(BaseCommand):
    """A simple management command which lists model fields and methods."""

    help = "List out the fields and methods for each model"

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--field-class",
            action="store_true",
            default=None,
            help="show class name of field.",
        )
        parser.add_argument(
            "--db-type",
            action="store_true",
            default=None,
            help="show database column type of field.",
        )
        parser.add_argument(
            "--signature",
            action="store_true",
            default=None,
            help="show the signature of method.",
        )
        parser.add_argument(
            "--all-methods",
            action="store_true",
            default=None,
            help="list all methods, including private and default.",
        )
        parser.add_argument(
            "--model",
            nargs="?",
            type=str,
            default=None,
            help="list the details for a single model. "
            "Input should be in the form appname.Modelname",
        )

    def list_model_info(self, options):
        style = color_style()
        INFO = getattr(style, "INFO", lambda x: x)
        WARN = getattr(style, "WARN", lambda x: x)
        BOLD = getattr(style, "BOLD", lambda x: x)

        FIELD_CLASS = (
            True
            if options.get("field_class", None) is not None
            else getattr(settings, "MODEL_INFO_FIELD_CLASS", False)
        )
        DB_TYPE = (
            True
            if options.get("db_type", None) is not None
            else getattr(settings, "MODEL_INFO_DB_TYPE", False)
        )
        SIGNATURE = (
            True
            if options.get("signature", None) is not None
            else getattr(settings, "MODEL_INFO_SIGNATURE", False)
        )
        ALL_METHODS = (
            True
            if options.get("all_methods", None) is not None
            else getattr(settings, "MODEL_INFO_ALL_METHODS", False)
        )
        MODEL = (
            options.get("model")
            if options.get("model", None) is not None
            else getattr(settings, "MODEL_INFO_MODEL", False)
        )

        default_methods = [
            "check",
            "clean",
            "clean_fields",
            "date_error_message",
            "delete",
            "from_db",
            "full_clean",
            "get_absolute_url",
            "get_deferred_fields",
            "prepare_database_save",
            "refresh_from_db",
            "save",
            "save_base",
            "serializable_value",
            "unique_error_message",
            "validate_unique",
        ]

        if MODEL:
            model_list = [django_apps.get_model(MODEL)]
        else:
            model_list = sorted(
                django_apps.get_models(),
                key=lambda x: (x._meta.app_label, x._meta.object_name),
                reverse=False,
            )
        for model in model_list:
            self.stdout.write(
                INFO(model._meta.app_label + "." + model._meta.object_name)
            )
            self.stdout.write(BOLD(HALFTAB + "Fields:"))

            for field in model._meta.get_fields():
                field_info = TAB + field.name + " -"

                if FIELD_CLASS:
                    try:
                        field_info += " " + field.__class__.__name__
                    except TypeError:
                        field_info += WARN(" TypeError (field_class)")
                    except AttributeError:
                        field_info += WARN(" AttributeError (field_class)")
                if FIELD_CLASS and DB_TYPE:
                    field_info += ","
                if DB_TYPE:
                    try:
                        field_info += " " + field.db_type(connection=connection)
                    except TypeError:
                        field_info += WARN(" TypeError (db_type)")
                    except AttributeError:
                        field_info += WARN(" AttributeError (db_type)")

                self.stdout.write(field_info)

            if ALL_METHODS:
                self.stdout.write(BOLD(HALFTAB + "Methods (all):"))
            else:
                self.stdout.write(BOLD(HALFTAB + "Methods (non-private/internal):"))

            for method_name in dir(model):
                try:
                    method = getattr(model, method_name)
                    if ALL_METHODS:
                        if callable(method) and not method_name[0].isupper():
                            if SIGNATURE:
                                signature = inspect.signature(method)
                            else:
                                signature = "()"
                            self.stdout.write(TAB + method_name + str(signature))
                    else:
                        if (
                            callable(method)
                            and not method_name.startswith("_")
                            and method_name not in default_methods
                            and not method_name[0].isupper()
                        ):
                            if SIGNATURE:
                                signature = inspect.signature(method)
                            else:
                                signature = "()"
                            self.stdout.write(TAB + method_name + str(signature))
                except AttributeError:
                    self.stdout.write(TAB + method_name + WARN(" - AttributeError"))
                except ValueError:
                    self.stdout.write(
                        TAB
                        + method_name
                        + WARN(" - ValueError (could not identify signature)")
                    )

            self.stdout.write("\n")

        self.stdout.write(INFO("Total Models Listed: %d" % len(model_list)))

    @signalcommand
    def handle(self, *args, **options):
        self.list_model_info(options)



# -*- coding: utf-8 -*-
from importlib import import_module
from inspect import (
    getmembers,
    isclass,
)
from pkgutil import walk_packages
from typing import (  # NOQA
    Dict,
    List,
    Tuple,
    Union,
)

from django.conf import settings
from django.utils.module_loading import import_string


class SubclassesFinder:
    def __init__(self, base_classes_from_settings):
        self.base_classes = []
        for element in base_classes_from_settings:
            if isinstance(element, str):
                element = import_string(element)
            self.base_classes.append(element)

    def _should_be_imported(self, candidate_to_import):  # type: (Tuple[str, type]) -> bool
        for base_class in self.base_classes:
            if issubclass(candidate_to_import[1], base_class):
                return True
        return False

    def collect_subclasses(self):  # type: () -> Dict[str, List[Tuple[str, str]]]
        """
        Collect all subclasses of user-defined base classes from project.
        :return: Dictionary from module name to list of tuples.
        First element of tuple is model name and second is alias.
        Currently we set alias equal to model name,
        but in future functionality of aliasing subclasses can be added.
        """
        result = {}  # type: Dict[str, List[Tuple[str, str]]]
        for loader, module_name, is_pkg in walk_packages(path=[str(settings.BASE_DIR)]):
            subclasses_from_module = self._collect_classes_from_module(module_name)
            if subclasses_from_module:
                result[module_name] = subclasses_from_module
        return result

    def _collect_classes_from_module(self, module_name):  # type: (str) -> List[Tuple[str, str]]
        for excluded_module in getattr(
            settings, "SHELL_PLUS_SUBCLASSES_IMPORT_MODULES_BLACKLIST", []
        ):
            if module_name.startswith(excluded_module):
                return []
        imported_module = import_module(module_name)
        classes_to_import = getmembers(
            imported_module,
            lambda element: isclass(element)
            and element.__module__ == imported_module.__name__,
        )
        classes_to_import = list(filter(self._should_be_imported, classes_to_import))
        return [(name, name) for name, _ in classes_to_import]



# -*- coding: utf-8 -*-
"""
Recreates the public schema for current database (PostgreSQL only).
Useful for Docker environments where you need to reset database
schema while there are active connections.
"""

import warnings

from django.core.management import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS
from django.db import connections
from django.conf import settings

from django_extensions.settings import POSTGRESQL_ENGINES
from django_extensions.utils.deprecation import RemovedInNextVersionWarning


class Command(BaseCommand):
    """`reset_schema` command implementation."""

    help = "Recreates the public schema for this project."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            default=True,
            help="Tells Django to NOT prompt the user for input of any kind.",
        )
        parser.add_argument(
            "-R",
            "--router",
            action="store",
            dest="router",
            default=DEFAULT_DB_ALIAS,
            help="Use this router-database instead of the one defined in settings.py",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to run command for. Defaults to the "%s".'
            % DEFAULT_DB_ALIAS,
        )
        parser.add_argument(
            "-S",
            "--schema",
            action="store",
            dest="schema",
            default="public",
            help='Drop this schema instead of "public"',
        )

    def handle(self, *args, **options):
        database = options["database"]
        if options["router"] != DEFAULT_DB_ALIAS:
            warnings.warn(
                "--router is deprecated. You should use --database.",
                RemovedInNextVersionWarning,
                stacklevel=2,
            )
            database = options["router"]

        dbinfo = settings.DATABASES.get(database)
        if dbinfo is None:
            raise CommandError("Unknown database %s" % database)

        engine = dbinfo.get("ENGINE")
        if engine not in POSTGRESQL_ENGINES:
            raise CommandError(
                "This command can be used only with PostgreSQL databases."
            )

        database_name = dbinfo["NAME"]

        schema = options["schema"]

        if options["interactive"]:
            confirm = input(
                """
You have requested a database schema reset.
This will IRREVERSIBLY DESTROY ALL data
in the "{}" schema of database "{}".
Are you sure you want to do this?

Type 'yes' to continue, or 'no' to cancel: """.format(schema, database_name)
            )
        else:
            confirm = "yes"

        if confirm != "yes":
            print("Reset cancelled.")
            return

        with connections[database].cursor() as cursor:
            cursor.execute("DROP SCHEMA {} CASCADE".format(schema))
            cursor.execute("CREATE SCHEMA {}".format(schema))



# -*- coding: utf-8 -*-
# Based on https://gist.github.com/voldmar/1264102
# and https://gist.github.com/runekaagaard/2eecf0a8367959dc634b7866694daf2c

import gc
import inspect
import weakref
from collections import defaultdict

import django
from django.apps import apps
from django.core.management.base import BaseCommand
from django.db.models.signals import (
    ModelSignal,
    pre_init,
    post_init,
    pre_save,
    post_save,
    pre_delete,
    post_delete,
    m2m_changed,
    pre_migrate,
    post_migrate,
)
from django.utils.encoding import force_str


MSG = "{module}.{name} #{line}{is_async}"

SIGNAL_NAMES = {
    pre_init: "pre_init",
    post_init: "post_init",
    pre_save: "pre_save",
    post_save: "post_save",
    pre_delete: "pre_delete",
    post_delete: "post_delete",
    m2m_changed: "m2m_changed",
    pre_migrate: "pre_migrate",
    post_migrate: "post_migrate",
}


class Command(BaseCommand):
    help = "List all signals by model and signal type"

    def handle(self, *args, **options):
        all_models = apps.get_models(include_auto_created=True, include_swapped=True)
        model_lookup = {id(m): m for m in all_models}

        signals = [obj for obj in gc.get_objects() if isinstance(obj, ModelSignal)]
        models = defaultdict(lambda: defaultdict(list))

        for signal in signals:
            signal_name = SIGNAL_NAMES.get(signal, "unknown")
            for receiver in signal.receivers:
                if django.VERSION >= (5, 0):
                    lookup, receiver, is_async = receiver
                else:
                    lookup, receiver = receiver
                    is_async = False
                if isinstance(receiver, weakref.ReferenceType):
                    receiver = receiver()
                if receiver is None:
                    continue
                receiver_id, sender_id = lookup

                model = model_lookup.get(sender_id, "_unknown_")
                if model:
                    models[model][signal_name].append(
                        MSG.format(
                            name=receiver.__name__,
                            module=receiver.__module__,
                            is_async=" (async)" if is_async else "",
                            line=inspect.getsourcelines(receiver)[1],
                            path=inspect.getsourcefile(receiver),
                        )
                    )

        output = []
        for key in sorted(models.keys(), key=str):
            verbose_name = force_str(key._meta.verbose_name)
            output.append(
                "{}.{} ({})".format(key.__module__, key.__name__, verbose_name)
            )
            for signal_name in sorted(models[key].keys()):
                lines = models[key][signal_name]
                output.append("    {}".format(signal_name))
                for line in lines:
                    output.append("        {}".format(line))

        return "\n".join(output)



# -*- coding: utf-8 -*-
import argparse
import string
import secrets
from typing import List

from django.core.management.base import BaseCommand
from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = "Generates a simple new password that can be used for a user password. "
    "Uses Python’s secrets module to generate passwords. Do not use this command to "
    "generate your most secure passwords."

    requires_system_checks: List[str] = []

    def add_arguments(self, parser):
        parser.add_argument(
            "-l", "--length", nargs="?", type=int, default=16, help="Password length."
        )
        parser.add_argument(
            "-c",
            "--complex",
            action=argparse.BooleanOptionalAction,
            help="More complex alphabet, includes punctuation",
        )

    @signalcommand
    def handle(self, *args, **options):
        length = options["length"]

        alphabet = string.ascii_letters + string.digits
        if options["complex"]:
            alphabet += string.punctuation
        return "".join(secrets.choice(alphabet) for i in range(length))



# -*- coding: utf-8 -*-
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from django_extensions.db.fields import (
    AutoSlugField,
    CreationDateTimeField,
    ModificationDateTimeField,
)


class TimeStampedModel(models.Model):
    """
    TimeStampedModel

    An abstract base class model that provides self-managed "created" and
    "modified" fields.
    """

    created = CreationDateTimeField(_("created"))
    modified = ModificationDateTimeField(_("modified"))

    def save(self, **kwargs):
        self.update_modified = kwargs.pop(
            "update_modified", getattr(self, "update_modified", True)
        )
        super().save(**kwargs)

    class Meta:
        get_latest_by = "modified"
        abstract = True


class TitleDescriptionModel(models.Model):
    """
    TitleDescriptionModel

    An abstract base class model that provides title and description fields.
    """

    title = models.CharField(_("title"), max_length=255)
    description = models.TextField(_("description"), blank=True, null=True)

    class Meta:
        abstract = True


class TitleSlugDescriptionModel(TitleDescriptionModel):
    """
    TitleSlugDescriptionModel

    An abstract base class model that provides title and description fields
    and a self-managed "slug" field that populates from the title.

    .. note ::
        If you want to use custom "slugify" function, you could
        define ``slugify_function`` which then will be used
        in :py:class:`AutoSlugField` to slugify ``populate_from`` field.

        See :py:class:`AutoSlugField` for more details.
    """

    slug = AutoSlugField(_("slug"), populate_from="title")

    class Meta:
        abstract = True


class ActivatorQuerySet(models.query.QuerySet):
    """
    ActivatorQuerySet

    Query set that returns statused results
    """

    def active(self):
        """Return active query set"""
        return self.filter(status=ActivatorModel.ACTIVE_STATUS)

    def inactive(self):
        """Return inactive query set"""
        return self.filter(status=ActivatorModel.INACTIVE_STATUS)


class ActivatorModelManager(models.Manager):
    """
    ActivatorModelManager

    Manager to return instances of ActivatorModel:
        SomeModel.objects.active() / .inactive()
    """

    def get_queryset(self):
        """Use ActivatorQuerySet for all results"""
        return ActivatorQuerySet(model=self.model, using=self._db)

    def active(self):
        """
        Return active instances of ActivatorModel:

        SomeModel.objects.active(), proxy to ActivatorQuerySet.active
        """
        return self.get_queryset().active()

    def inactive(self):
        """
        Return inactive instances of ActivatorModel:

        SomeModel.objects.inactive(), proxy to ActivatorQuerySet.inactive
        """
        return self.get_queryset().inactive()


class ActivatorModel(models.Model):
    """
    ActivatorModel

    An abstract base class model that provides activate and deactivate fields.
    """

    INACTIVE_STATUS = 0
    ACTIVE_STATUS = 1

    STATUS_CHOICES = (
        (INACTIVE_STATUS, _("Inactive")),
        (ACTIVE_STATUS, _("Active")),
    )
    status = models.IntegerField(
        _("status"), choices=STATUS_CHOICES, default=ACTIVE_STATUS
    )
    activate_date = models.DateTimeField(
        blank=True, null=True, help_text=_("keep empty for an immediate activation")
    )
    deactivate_date = models.DateTimeField(
        blank=True, null=True, help_text=_("keep empty for indefinite activation")
    )
    objects = ActivatorModelManager()

    class Meta:
        ordering = (
            "status",
            "-activate_date",
        )
        abstract = True

    def save(self, *args, **kwargs):
        if not self.activate_date:
            self.activate_date = now()
        super().save(*args, **kwargs)



# -*- coding: utf-8 -*-
"""
Make debugging Django templates easier.

Example:

    {% load debugger_tags %}

    {{ object|ipdb }}

"""

from django import template


register = template.Library()


@register.filter
def ipdb(obj):  # pragma: no cover
    """Interactive Python debugger filter."""
    __import__("ipdb").set_trace()
    return obj


@register.filter
def pdb(obj):
    """Python debugger filter."""
    __import__("pdb").set_trace()
    return obj


@register.filter
def wdb(obj):  # pragma: no cover
    """Web debugger filter."""
    __import__("wdb").set_trace()
    return obj



# -*- coding: utf-8 -*-
import time
import logging
from hashlib import md5

# also see: https://djangosnippets.org/snippets/2242/


class RateLimiterFilter(logging.Filter):
    def filter(self, record):
        from django.conf import settings
        from django.core.cache import cache

        # Rate is specified as 1 messages logged per N seconds. (aka cache timeout)
        rate = getattr(settings, "RATE_LIMITER_FILTER_RATE", 10)
        prefix = getattr(settings, "RATE_LIMITER_FILTER_PREFIX", "ratelimiterfilter")

        subject = record.getMessage()
        cache_key = "%s:%s" % (prefix, md5(subject).hexdigest())
        cache_count_key = "%s:count" % cache_key

        result = cache.get_many([cache_key, cache_count_key])
        value = result.get(cache_key)
        cntr = result.get(cache_count_key)

        if not cntr:
            cntr = 1
            cache.set(cache_count_key, cntr, rate + 60)

        if value:
            cache.incr(cache_count_key)
            return False

        record.msg = "[%sx] %s" % (cntr, record.msg)
        cache.set(cache_key, time.time(), rate)
        return True
